<?php
/**
 *  LDAP.php
 *
 *  PHP versions 4 and 5
 *
 *  @package    Auth
 *
 *  @author     Kaoru Sekiguchi <sekiguchi.kaoru@secioss.co.jp>
 *  @copyright  2016 SECIOSS, INC.
 *
 *  @version    CVS: $Id$
 */
require_once 'Secioss/AutoLogin/Container.php';
require_once 'HTTP/Request.php';

define('PASSWORD_SYNTAX', '2.5.4.35');

/**
 *  AutoLogin_Container_LDAP
 *
 *  @package    Auth
 *
 *  @author     Kaoru Sekiguchi <sekiguchi.kaoru@secioss.co.jp>
 *  @copyright  2016 SECIOSS, INC.
 *
 *  @version    CVS: $Id$
 */
class AutoLogin_Container_LDAP extends AutoLogin_Container
{
    // {{{ AutoLogin_Container_LDAP

    /**
     *  AutoLogin_Container_LDAPクラスのコンストラクタ
     *
     *  @access public
     *
     *  @param  mixed   $options        LDAPの設定
     *
     *  @return mixed   0:正常終了 PEAR_Error:エラー
     */
    public function __construct($options)
    {
        parent::__construct($options);
        if ($this->options['pwhash'] == 'AD') {
            $this->options['basedn'] = mb_convert_encoding($this->options['basedn'], 'SJIS');
        }

        $rc = 0;
        if (strtoupper(substr(PHP_OS, 0, 3)) != 'WIN') {
            $rc = $this->_connect();
            register_shutdown_function([&$this, '_disconnect']);
        }

        return $rc;
    }

    // }}}

    // {{{ _connect()

    /**
     * LDAPサーバーに接続する
     *
     * @access private
     *
     * @return mixed 0:正常終了 PEAR_Error:エラー
     */
    public function _connect()
    {
        $this->conn = @ldap_connect($this->options['uri']);
        ldap_set_option($this->conn, LDAP_OPT_PROTOCOL_VERSION, 3);
        if (!@ldap_bind($this->conn, $this->options['binddn'], $this->options['bindpw'])) {
            return PEAR::raiseError(ldap_error($this->conn), ldap_errno($this->conn));
        }

        return 0;
    }

    // }}}

    // {{{ _disconnect()

    /**
     * LDAPサーバーへの接続を切断する
     *
     * @access private
     */
    public function _disconnect()
    {
        @ldap_unbind($this->conn);
    }

    // {{{ fetchData()

    /**
     * LDAPサーバーからユーザー情報を取得する
     *
     * @param  string   ユーザー名
     * @param mixed $username
     *
     * @return mixed true:正常終了 PEAR_Error:エラー
     */
    public function fetchData($username)
    {
        if (!$username || strlen($username) > 255) {
            return PEAR::raiseError('Invalid user id', AUTO_LOGIN_INVALID_VALUE);
        }

        $tenant = '';
        if (preg_match('/^[^@]+@(.*)$/', $username, $matches)) {
            $tenant = $matches[1];
            $tenant = preg_replace('/([*()])/', '\\$1', $tenant);
        }

        if (strtoupper(substr(PHP_OS, 0, 3)) == 'WIN') {
            $this->_connect();
        }

        $filter = '('.$this->options['userattr'].'='.preg_replace('/([*()])/', '\\$1', $username).')';
        if ($this->options['userfilter']) {
            $filter = '(&'.$filter.$this->options['userfilter'].')';
        }
        $res = @ldap_search($this->conn, $this->options['basedn'], $filter);
        if ($res == false) {
            return PEAR::raiseError(ldap_error($this->conn), ldap_errno($this->conn));
        }

        $num = ldap_count_entries($this->conn, $res);

        if ($num == 0) {
            return PEAR::raiseError("User doesn't exist", AUTO_LOGIN_NO_USER);
        } elseif ($num != 1) {
            return PEAR::raiseError("User isn't unique", AUTO_LOGIN_ERROR);
        }

        $entries = ldap_get_entries($this->conn, $res);
        $entry = $entries[0];

        if (strtoupper(substr(PHP_OS, 0, 3)) == 'WIN') {
            $this->_disconnect();
        }

        $this->id = $entry['dn'];

        $this->prop = [];
        for ($i = 0; $i < $entry['count']; $i++) {
            $key = $entry[$i];

            if ($entry[$key]['count'] == 1) {
                $this->prop[$key] = $entry[$key][0];
            } else {
                $v = [];
                for ($j = 0; $j < $entry[$key]['count']; $j++) {
                    $v[] = $entry[$key][$j];
                }
                $this->prop[$key] = $v;
            }
        }

        if (!isset($this->prop['seciossloginid'])) {
            $this->prop['seciossloginid'] = $username;
        }
        $this->prop['seciosssystemid'] = preg_replace('/@[^@]+$/', '', $username);

        if (isset($this->prop[$this->options['loginattr']])) {
            $loginid = $this->prop[$this->options['loginattr']];
            if (is_array($loginid)) {
                $this->loginid = $loginid[0];
            } else {
                $this->loginid = $loginid;
            }
        }

        if (isset($this->prop[$this->options['statusattr']])) {
            $this->status = $this->prop[$this->options['statusattr']];
        }

        if (isset($this->prop[$this->options['pwdlockoutattr']])) {
            if ($this->prop[$this->options['pwdlockoutattr']] == 'TRUE') {
                $this->pwdlockout = true;
            } else {
                $this->pwdlockout = false;
            }
        }

        if (!$tenant && preg_match('/[^,]+,[^,]+,'.$this->options['tenantattr'].'=([^,]+),'.$this->options['basedn'].'/i', $this->id, $matches)) {
            $tenant = $matches[1];
        }

        if ($this->options['pwdpolicydb']) {
            $rc = $this->_getPwdPolicy($tenant);
            if (PEAR::isError($rc)) {
                return $rc;
            }
        }

        return true;
    }

    // }}}

    // {{{ auth()

    /**
     *  ユーザーの認証を行う
     *
     *  @access public
     *
     *  @param  string   パスワード
     * @param mixed $password
     *
     *  @return string   0:成功 1:失敗
     */
    public function auth($password)
    {
        if (!$password) {
            return 1;
        }

        if ($this->options['oldpwdauth'] && isset($this->prop[$this->options['pwdmustchangeattr']]) && $this->prop[$this->options['pwdmustchangeattr']] == 'TRUE') {
            if (isset($this->prop[$this->options['pwdhistoryattr']])) {
                $pwdhistory = $this->_to_array($this->prop[$this->options['pwdhistoryattr']]);
                if (!preg_match('/#([^#]+)$/', $pwdhistory[0], $matches)) {
                    return 1;
                }
                $oldpasswd = $matches[1];

                $hash = '';
                if (preg_match('/^{([^}]+)}/', $oldpasswd, $matches)) {
                    $hash = $matches[1];
                }

                if ($this->cmpPasswd($password, $oldpasswd, $hash)) {
                    return 0;
                }
            }
        }

        if (strtoupper(substr(PHP_OS, 0, 3)) == 'WIN') {
            $this->_connect();
        }

        $rc = @ldap_bind($this->conn, $this->id, $password) ? 0 : 1;
        @ldap_bind($this->conn, $this->options['binddn'], $this->options['bindpw']);

        if (strtoupper(substr(PHP_OS, 0, 3)) == 'WIN') {
            $this->_disconnect();
        }

        return $rc;
    }

    // }}}

    // {{{ getPassword()

    /**
     *  暗号化されているパスワードを復号化して取得する
     *
     *  @access public
     *
     * @param null|mixed $app
     *
     *  @return string   パスワード
     */
    public function getPassword($app = null)
    {
        $password = '';
        $regex_app = addcslashes($app, '.*+?[]()|^$\\/');

        if (isset($this->prop[$this->options['encryptpwdattr']])) {
            $encryptpwds = $this->_to_array($this->prop[$this->options['encryptpwdattr']]);

            for ($i = 0; $i < count($encryptpwds); $i++) {
                if ($app) {
                    if (preg_match('/^{'.$regex_app.'}(.+)/i', $encryptpwds[$i], $matches)) {
                        $password = $this->decrypt($matches[1]);
                        break;
                    }
                }
                if (!preg_match('/^{[^}]+}/', $encryptpwds[$i])) {
                    $password = $this->decrypt($encryptpwds[$i]);
                }
            }
        }

        return $password;
    }

    // }}}

    public function getTenants($tenant = null)
    {
        if ($tenant) {
            $tenant = addcslashes($tenant, '*()`\\');
            $filter = '('.$this->options['tenantattr']."=$tenant)";
        } else {
            $filter = '('.$this->options['tenantattr'].'=*)';
        }
        if ($this->options['tenantfilter']) {
            $filter = '(&'.$filter.$this->options['tenantfilter'].')';
        }
        $res = @ldap_list($this->conn, $this->options['basedn'], $filter, [$this->options['tenantattr'], 'seciossallowedfunction', 'seciossconfigserializeddata', 'seciossconfigserializeddata;x-type-html', 'mail', 'seciosstenantmaxusers;x-type-device']);
        if ($res == false) {
            return PEAR::raiseError(ldap_error($this->conn), ldap_errno($this->conn));
        }

        $entries = ldap_get_entries($this->conn, $res);
        $tenants = [];
        for ($i = 0; $i < $entries['count']; $i++) {
            $tenants[$entries[$i][$this->options['tenantattr']][0]] = ['func' => isset($entries[$i]['seciossallowedfunction']) ? $entries[$i]['seciossallowedfunction'] : []];
            if (isset($entries[$i]['seciossconfigserializeddata'])) {
                $tenantconf = unserialize($entries[$i]['seciossconfigserializeddata'][0]);
                if ($tenantconf) {
                    foreach (array_keys($tenantconf) as $key) {
                        if (!is_array($tenantconf[$key])) {
                            continue;
                        }
                        $tenants[$entries[$i][$this->options['tenantattr']][0]][$key] = $tenantconf[$key];
                    }
                }
            }
            if (isset($entries[$i]['seciossconfigserializeddata;x-type-html'])) {
                $tpl_html = unserialize($entries[$i]['seciossconfigserializeddata;x-type-html'][0]);
                if ($tpl_html) {
                    foreach (array_keys($tpl_html) as $key) {
                        $tenants[$entries[$i][$this->options['tenantattr']][0]]['template']['v2'][$key] = $tpl_html[$key];
                    }
                }
            }

            if (isset($entries[$i]['mail'])) {
                $tenants[$entries[$i][$this->options['tenantattr']][0]]['mail'] = $entries[$i]['mail'];
            }
            if (isset($entries[$i]['mail;x-type-admin'])) {
                $tenants[$entries[$i][$this->options['tenantattr']][0]]['mail;x-type-admin'] = $entries[$i]['mail;x-type-admin'];
            }
            if (isset($entries[$i]['seciosstenantmaxusers;x-type-device'])) {
                $tenants[$entries[$i][$this->options['tenantattr']][0]]['seciosstenantmaxusers;x-type-device'] = $entries[$i]['seciosstenantmaxusers;x-type-device'][0];
            }
        }

        return $tenants;
    }

    public function getDeviceCurrentNum($ltype, $dtype, $newid, $tenant)
    {
        $num = 0;
        $ltype_list = ['device', 'user', 'deviceperuser'];
        if (!$ltype || !in_array($ltype, $ltype_list)) {
            return PEAR::raiseError('Invalid license type', AUTO_LOGIN_INVALID_VALUE);
        }
        if (!$dtype) {
            return PEAR::raiseError('Invalid device type', AUTO_LOGIN_INVALID_VALUE);
        }
        if (!$newid) {
            return PEAR::raiseError('Invalid device id', AUTO_LOGIN_INVALID_VALUE);
        }

        $basedn = $this->options['basedn'];
        if ($tenant) {
            $basedn = $this->options['tenantattr']."=$tenant,$basedn";
        }
        $basedn = "ou=people,$basedn";
        $filter = '(seciossdeviceid=*)';
        if ($this->options['userfilter']) {
            $filter = '(&'.$filter.$this->options['userfilter'].')';
        }
        if ($this->options['statusattr']) {
            $filter = '(&'.$filter.'(!('.$this->options['statusattr'].'=deleted)))';
        }
        $attrs = [];
        if ((strcmp('smartphone', $dtype) === 0) || (strcmp('computer', $dtype) === 0)) {
            array_push($attrs, 'seciossdeviceid;x-dev-smartphone');
            array_push($attrs, 'seciossdeviceid;x-dev-computer');
        } else {
            array_push($attrs, 'seciossdeviceid;x-dev-'.$dtype);
        }
        $keys = array_merge(['dn'], $attrs);
        $res = @ldap_list($this->conn, $basedn, $filter, $keys);
        if ($res == false) {
            return PEAR::raiseError(ldap_error($this->conn), ldap_errno($this->conn));
        }
        $users = ldap_get_entries($this->conn, $res);
        if (strcmp('user', $ltype) === 0) {
            for ($i = 0; $i < $users['count'] && isset($users[$i]); $i++) {
                foreach ($attrs as $attr) {
                    if (isset($users[$i][$attr])) {
                        if (strcasecmp($this->id, $users[$i]['dn']) === 0) {
                            break;
                        }
                        $ids = $users[$i][$attr];
                        $exist = false;
                        for ($j = 0; $j < $ids['count'] && isset($ids[$j]); $j++) {
                            if (!preg_match('/^[^#]+#active/', $ids[$j])) {
                                $exist = true;
                                break;
                            }
                        }
                        if ($exist) {
                            $num++;
                            break;
                        }
                    }
                }
            }
        } elseif (strcmp('deviceperuser', $ltype) === 0) {
            for ($i = 0; $i < $users['count'] && isset($users[$i]); $i++) {
                if (strcasecmp($this->id, $users[$i]['dn']) === 0) {
                    foreach ($attrs as $attr) {
                        if (isset($users[$i][$attr])) {
                            $ids = $users[$i][$attr];
                            for ($j = 0; $j < $ids['count'] && isset($ids[$j]); $j++) {
                                if (!preg_match('/^[^#]+#active/', $ids[$j])) {
                                    continue;
                                }
                                if (preg_match('/^'.$newid.'#[^#]+#?([^#]*)/', $ids[$j])) {
                                    continue;
                                }
                                $num++;
                            }
                        }
                    }
                    break;
                }
            }
        } elseif (strcmp('device', $ltype) === 0) {
            for ($i = 0; $i < $users['count'] && isset($users[$i]); $i++) {
                foreach ($attrs as $attr) {
                    if (isset($users[$i][$attr])) {
                        $ids = $users[$i][$attr];
                        $dn = $users[$i]['dn'];
                        for ($j = 0; $j < $ids['count'] && isset($ids[$j]); $j++) {
                            if (!preg_match('/^[^#]+#active/', $ids[$j])) {
                                continue;
                            }
                            if ((strcasecmp($this->id, $dn) === 0) && (preg_match('/^'.$newid.'#[^#]+#?([^#]*)/', $ids[$j]))) {
                                continue;
                            }
                            $num++;
                        }
                    }
                }
            }
        }
        return $num;
    }

    // {{{ setProp

    /**
     *  プロパティへのアクセサ(W)
     *
     *  @access public
     *
     *  @param  array   $prop    プロパティ
     *
     *  @return mixed   true:正常終了 PEAR_Error:エラー
     */
    public function setProp($prop)
    {
        if (!$this->id) {
            return PEAR::raiseError('Must fetch data', AUTO_LOGIN_ERROR);
        }

        if ($this->options['api_url']) {
            $rc = $this->_rest_modify($this->id, $prop);
            if (PEAR::isError($rc)) {
                return $rc;
            }
        } else {
            if (strtoupper(substr(PHP_OS, 0, 3)) == 'WIN') {
                $this->_connect();
            }

            if (!@ldap_mod_replace($this->conn, $this->id, $prop)) {
                return PEAR::raiseError(ldap_error($this->conn), ldap_errno($this->conn));
            }

            if (strtoupper(substr(PHP_OS, 0, 3)) == 'WIN') {
                $this->_disconnect();
            }
        }

        return true;
    }

    // }}}

    // {{{ deleteProp

    /**
     *  プロパティへのアクセサ(W)
     *
     *  @access public
     *
     *  @param  array   $prop    プロパティ
     *
     *  @return mixed   true:正常終了 PEAR_Error:エラー
     */
    public function delProp($prop)
    {
        if (!$this->id) {
            return PEAR::raiseError('Must fetch data', AUTO_LOGIN_ERROR);
        }

        if ($this->options['api_url']) {
            $rc = $this->_rest_modify($this->id, $prop, 'delete');
            if (PEAR::isError($rc)) {
                return $rc;
            }
        } else {
            if (strtoupper(substr(PHP_OS, 0, 3)) == 'WIN') {
                $this->_connect();
            }

            if (!@ldap_mod_del($this->conn, $this->id, $prop)) {
                return PEAR::raiseError(ldap_error($this->conn), ldap_errno($this->conn));
            }

            if (strtoupper(substr(PHP_OS, 0, 3)) == 'WIN') {
                $this->_disconnect();
            }
        }

        return true;
    }

    // }}}

    // {{{ setPassword()

    /**
     *  パスワードを暗号化してLDAPに格納する
     *
     *  @access public
     *
     *  @param  string   パスワード
     * @param null|mixed $password
     * @param null|mixed $app
     * @param mixed      $init
     * @param null|mixed $random
     *
     *  @return mixed    true:正常終了 PEAR_Error:エラー
     */
    public function setPassword($password = null, $app = null, $init = false, $random = null)
    {
        $regex_app = addcslashes($app, '.*+?[]()|^$\\/');

        if (!$this->id) {
            return PEAR::raiseError('Must fetch data', AUTO_LOGIN_ERROR);
        }

        if (!$password || strlen($password) > 255) {
            return PEAR::raiseError('Invalid password', AUTO_LOGIN_INVALID_VALUE);
        }

        if ($this->options['encryptpwdattr']) {
            $encrypt = $this->encrypt($password);
            if (PEAR::isError($encrypt)) {
                return $encrypt;
            }

            $encryptpwds = [];
            if (isset($this->prop[$this->options['encryptpwdattr']])) {
                $encryptpwds = $this->_to_array($this->prop[$this->options['encryptpwdattr']]);
            }
        }

        if ($this->options['randompwdattr']) {
            if (!$random) {
                $pwlen = strlen($password);
                $random = $this->random($pwlen);
            }
            $rndencrypt = $this->encrypt($random);
            if (PEAR::isError($rndencrypt)) {
                return $rndencrypt;
            }

            $info[$this->options['randompwdattr']] = $rndencrypt;
        }

        if (!$app) {
            if ($this->options['pwdattr']) {
                $hash = $this->options['pwhash'];
                $tab = '';
                if ($hash == 'CRYPT' || $hash == 'MD5' || $hash == 'SHA' || $hash == 'SSHA') {
                    $tab = '{'.$hash.'}';
                }
                $info[$this->options['pwdattr']] = $tab.$this->hashPasswd($password);
            }

            if ($this->options['encryptpwdattr']) {
                $match = false;
                for ($i = 0; $i < count($encryptpwds); $i++) {
                    if (!preg_match('/^{[^}]+}/', $encryptpwds[$i])) {
                        $encryptpwds[$i] = $encrypt;
                        $match = true;
                    }
                }
                if (!$match) {
                    array_push($encryptpwds, $encrypt);
                }
                $info[$this->options['encryptpwdattr']] = $encryptpwds;
            }

            $oldpwdtime = '';
            if ($this->options['pwdtimeattr']) {
                $info[$this->options['pwdtimeattr']] = date('YmdHis').'Z';
                if (isset($this->prop[$this->options['pwdtimeattr']])) {
                    $oldpwdtime = $this->prop[$this->options['pwdtimeattr']];
                }
            }

            if ($this->options['pwdhistoryattr']) {
                $oldpasswd = isset($this->prop[$this->options['pwdattr']]) ? $this->prop[$this->options['pwdattr']] : '';
                $oldpwdtime = $oldpwdtime ? $oldpwdtime : '19700101000000Z';

                if ($oldpasswd) {
                    $pwdhistory = [];
                    $oldpasswd = $oldpwdtime.'#'.PASSWORD_SYNTAX.'#'.strlen($oldpasswd).'#'.$oldpasswd;
                    $pwinhistory = $this->options['pwinhistory'];
                    if ($pwinhistory) {
                        if (isset($this->prop[$this->options['pwdhistoryattr']])) {
                            $pwdhistory = $this->_to_array($this->prop[$this->options['pwdhistoryattr']]);
                        }

                        if (count($pwdhistory) >= $pwinhistory - 1) {
                            for ($i = 0; $i < count($pwdhistory) - $pwinhistory + 1; $i++) {
                                array_pop($pwdhistory);
                            }
                        }
                    }
                    array_unshift($pwdhistory, $oldpasswd);

                    $info[$this->options['pwdhistoryattr']] = $pwdhistory;
                }
            }

            if ($this->options['pwdmustchangeattr']) {
                $info[$this->options['pwdmustchangeattr']] = $init ? 'TRUE' : 'FALSE';
                $info[$this->options['pwdexpwarnedattr']] = 'FALSE';
                $info[$this->options['pwdexpiredattr']] = 'FALSE';
            }
        } else {
            if ($this->options['encryptpwdattr']) {
                $match = false;
                for ($i = 0; $i < count($encryptpwds); $i++) {
                    if (preg_match('/^{'.$regex_app.'}/i', $encryptpwds[$i])) {
                        $encryptpwds[$i] = '{'.$app."}$encrypt";
                        $match = true;
                    }
                }
                if (!$match) {
                    array_push($encryptpwds, '{'.$app."}$encrypt");
                }
                $info[$this->options['encryptpwdattr']] = $encryptpwds;
            }
        }

        if ($this->options['api_url']) {
            $rc = $this->_rest_modify($this->id, $info);
            if (PEAR::isError($rc)) {
                return $rc;
            }
        } else {
            if (strtoupper(substr(PHP_OS, 0, 3)) == 'WIN') {
                $this->_connect();
            }

            if (!@ldap_mod_replace($this->conn, $this->id, $info)) {
                if (ldap_errno($this->conn) == 19) {
                    return PEAR::raiseError(ldap_error($this->conn), AUTO_LOGIN_INVALID_VALUE);
                } else {
                    return PEAR::raiseError(ldap_error($this->conn), ldap_errno($this->conn));
                }
            }

            if (strtoupper(substr(PHP_OS, 0, 3)) == 'WIN') {
                $this->_disconnect();
            }
        }

        return true;
    }

    // }}}

    public function deletePassword($app)
    {
        if (!$app) {
            return true;
        }
        $regex_app = addcslashes($app, '.*+?[]()|^$\\/');

        if (isset($this->prop[$this->options['encryptpwdattr']])) {
            $encryptpwds = $this->_to_array($this->prop[$this->options['encryptpwdattr']]);

            for ($i = 0; $i < count($encryptpwds); $i++) {
                if (preg_match('/^{'.$regex_app.'}/i', $encryptpwds[$i])) {
                    if (strtoupper(substr(PHP_OS, 0, 3)) == 'WIN') {
                        $this->_connect();
                    }

                    if (!@ldap_mod_del($this->conn, $this->id, [$this->options['encryptpwdattr'] => $encryptpwds[$i]])) {
                        return PEAR::raiseError(ldap_error($this->conn), ldap_errno($this->conn));
                    }

                    if (strtoupper(substr(PHP_OS, 0, 3)) == 'WIN') {
                        $this->_disconnect();
                    }

                    return true;
                }
            }
        }

        return true;
    }

    // {{{ getPwdChangedTime()

    /**
     *  パスワード変更時間を取得する
     *
     *  @access public
     *
     *  @return string   パスワード変更時間（秒）
     */
    public function getPwdChangedTime()
    {
        if (!$this->options['pwdtimeattr']) {
            return 0;
        }

        $pwdtime = $this->prop[$this->options['pwdtimeattr']];

        if (!$pwdtime) {
            return 0;
        }

        if (!preg_match('/^([0-9]{4})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})([0-9]{2})Z$/', $pwdtime, $matches)) {
            return 0;
        }

        return mktime($matches[4], $matches[5], $matches[6], $matches[2], $matches[3], $matches[1]);
    }

    // }}}

    // {{{ getPwdHistory()

    /**
     *  パスワード履歴を取得する
     *
     *  @access public
     *
     *  @return mixed   パスワード履歴
     */
    public function getPwdHistory()
    {
        if (!$this->options['pwdhistoryattr']) {
            return [];
        }

        if (!isset($this->prop[$this->options['pwdhistoryattr']])) {
            $this->prop[$this->options['pwdhistoryattr']] = [];
        }

        $pwdhistory = $this->_to_array($this->prop[$this->options['pwdhistoryattr']]);

        for ($i = 0; $i < count($pwdhistory); $i++) {
            $pwdhistory[$i] = preg_replace('/^[^#]+#[^#]+#[^#]+#/', '', $pwdhistory[$i]);
        }
        array_unshift($pwdhistory, $this->prop[$this->options['pwdattr']]);

        return $pwdhistory;
    }

    // }}}

    public function isPwdMustChange()
    {
        if (isset($this->prop[$this->options['pwdmustchangeattr']]) &&
            $this->prop[$this->options['pwdmustchangeattr']] == 'TRUE') {
            return true;
        }
        return false;
    }

    // {{{ getProfiles()

    /**
     * プロファイル一覧を取得する
     *
     * @access public
     *
     * @return array プロファイルDN一覧
     */
    public function getProfiles()
    {
        $profiles = [];

        $profiles = array_merge($profiles, (array) $this->get('seciossbusinessrole'));
        $profiles = array_merge($profiles, (array) $this->get('seciossbusinessrole;x-perm-group'));

        return $profiles;
    }

    // }}}

    // {{{ getProfileConf($tenant)

    /**
     * プロファイルのカスタムコンフィグを取得する
     *
     * @access public
     *
     * @param null|mixed $tenant
     *
     * @return array プロファイルのカスタムコンフィグ
     */
    public function getProfileConf($tenant = null)
    {
        $profileconf = [];

        if (strtoupper(substr(PHP_OS, 0, 3)) == 'WIN') {
            $this->_connect();
        }

        $filter = null;
        $profiles = $this->getProfiles();
        foreach ($profiles as $profile) {
            if (preg_match('/^([^=]+=[^,]+),/', $profile, $matches)) {
                if ($filter) {
                    $filter = '(|('.$matches[1].')'.$filter.')';
                } else {
                    $filter = '('.$matches[1].')';
                }
            }
        }
        if (!$filter) {
            return $profileconf;
        }

        $filter = "(&(&(objectClass=seciossRole)(seciossConfigSerializedData=*))$filter)";
        $basedn = $tenant ? $this->options['tenantattr']."=$tenant,".$this->options['basedn'] : $this->options['basedn'];
        $res = @ldap_search($this->conn, $basedn, $filter);
        if ($res == false) {
            return PEAR::raiseError(ldap_error($this->conn), ldap_errno($this->conn));
        }

        $entries = ldap_get_entries($this->conn, $res);
        for ($i = 0; $i < $entries['count']; $i++) {
            $seciossconfigserializeddata = $entries[$i]['seciossconfigserializeddata'][0];
            $tmpcpnf = unserialize($seciossconfigserializeddata);
            $profileconf = hashMerge($profileconf, $tmpcpnf);
        }

        return $profileconf;
    }

    // }}}

    // {{{ _getPwdPolicy($tenant)
    public function _getPwdPolicy($tenant = null)
    {
        $pwdpolicy = null;

        if (strtoupper(substr(PHP_OS, 0, 3)) == 'WIN') {
            $this->_connect();
        }

        $filter = null;
        $profiles = $this->getProfiles();
        foreach ($profiles as $profile) {
            if (preg_match('/^([^=]+=[^,]+),/', $profile, $matches)) {
                if ($filter) {
                    $filter = '(|('.$matches[1].')'.$filter.')';
                } else {
                    $filter = '('.$matches[1].')';
                }
            }
        }
        if ($filter) {
            $filter = "(&(&(objectClass=seciossRole)(seciossPwdPolicyEnabled=TRUE))$filter)";
            $res = @ldap_search($this->conn, $tenant ? $this->options['tenantattr']."=$tenant,".$this->options['basedn'] : $this->options['basedn'], $filter);
            if ($res == false) {
                return PEAR::raiseError(ldap_error($this->conn), ldap_errno($this->conn));
            }

            $priority = 0;
            $entries = ldap_get_entries($this->conn, $res);
            for ($i = 0; $i < $entries['count']; $i++) {
                if ($entries[$i]['seciossrolespecification'][0] > $priority) {
                    $pwdpolicy = $entries[$i];
                    $priority = $entries[$i]['seciossrolespecification'][0];
                }
            }
        }

        if (!$pwdpolicy && $tenant) {
            $res = @ldap_search($this->conn, $this->options['basedn'], '(&(objectClass=organization)(&(seciossPwdPolicyEnabled=TRUE)('.$this->options['tenantattr']."=$tenant)))");
            if ($res == false) {
                return PEAR::raiseError(ldap_error($this->conn), ldap_errno($this->conn));
            }

            $num = ldap_count_entries($this->conn, $res);
            if ($num > 0) {
                $entries = ldap_get_entries($this->conn, $res);
                $pwdpolicy = $entries[0];
            }
        }

        if (strtoupper(substr(PHP_OS, 0, 3)) == 'WIN') {
            $this->_disconnect();
        }

        if (!$pwdpolicy) {
            return true;
        }

        $this->options['pwminlen'] = 0;
        $this->options['pwmaxlen'] = 255;
        $this->options['pwallow'] = null;
        $this->options['pwallowlimit'] = 0;
        $this->options['pwdeny'] = null;
        $this->options['pwminage'] = 0;
        $this->options['pwinhistory'] = 0;
        $this->options['pwstrength'] = 0;
        $this->options['pwprohibitattr'] = null;

        for ($i = 0; $i < $pwdpolicy['count']; $i++) {
            $key = $pwdpolicy[$i];
            if ($pwdpolicy[$key]['count'] == 1) {
                $v = $pwdpolicy[$key][0];
            } else {
                $v = [];
                for ($j = 0; $j < $pwdpolicy[$key]['count']; $j++) {
                    $v[] = $pwdpolicy[$key][$j];
                }
            }
            if (!is_array($v) && preg_match('/^ *$/', $v)) {
                continue;
            }

            switch ($key) {
                case 'pwdminlength':
                    $this->options['pwminlen'] = $v;
                    break;
                case 'seciosspwdmaxlength':
                    $this->options['pwmaxlen'] = $v;
                    break;
                case 'seciosspwdallowedchars':
                    $v = is_array($v) ? $v : [$v];
                    $this->options['pwallow'] = $v;
                    break;
                case 'seciosspwddeniedchars':
                    $v = is_array($v) ? $v : [$v];
                    $this->options['pwdeny'] = $v;
                    break;
                case 'pwdminage':
                    $this->options['pwminage'] = $v;
                    break;
                case 'pwdinhistory':
                    $this->options['pwinhistory'] = $v;
                    break;
                case 'pwdcheckquality':
                    $this->options['pwstrength'] = $v;
                    break;
                case 'seciosspwdserializeddata':
                    $v = unserialize($v);
                    $this->options['pwprohibitattr'] = isset($v['pwprohibitattr']) ? implode(',', $v['pwprohibitattr']) : null;
                    $this->options['pwallowlimit'] = isset($v['pwallowlimit']) ? $v['pwallowlimit'] : 0;
                    break;
            }
        }

        return true;
    }

    // }}}

    // {{{ _setDefaults()

    /**
     * optionsにデフォルト値を設定する
     *
     * @access private
     */
    public function _setDefaults()
    {
        parent::_setDefaults();

        $this->options['uri'] = 'ldap://localhost';
        $this->options['binddn'] = '';
        $this->options['bindpw'] = '';
        $this->options['basedn'] = '';
        $this->options['userattr'] = 'uid';
        $this->options['pwdattr'] = 'userpassword';
        $this->options['loginattr'] = 'uid';
        $this->options['statusattr'] = 'seciossaccountstatus';
        $this->options['encryptpwdattr'] = 'seciossencryptedpassword';
        $this->options['randompwdattr'] = 'seciossrandompassword';
        $this->options['pwdtimeattr'] = '';
        $this->options['tenantattr'] = 'o';
        $this->options['tenantfilter'] = '(&(objectClass=organization)(!(seciossTenantStatus=inactive)))';
        $this->options['pwdhistoryattr'] = '';
        $this->options['pwdmustchangeattr'] = 'seciosspwdmustchange';
        $this->options['pwdexpwarnedattr'] = 'seciosspwdexpwarned';
        $this->options['pwdexpiredattr'] = 'seciosspwdexpired';
        $this->options['pwdlockoutattr'] = 'seciosspwdlockout';
        $this->options['pwdpolicydb'] = '';
        $this->options['secretattr'] = 'seciossotpinitsecret';
        $this->options['userfilter'] = '';
        $this->options['oldpwdauth'] = false;
        $this->options['api_url'] = '';
        $this->options['api_basedn'] = '';
    }

    // }}}

    /**
     * seciossDeviceId;x-dev-$device: $deviceid#$status=$os+DATE+$ip#user#note#appver
     *
     * @param mixed $device
     * @param mixed $deviceid
     * @param mixed $status
     * @param mixed $os
     * @param mixed $ip
     * @param mixed $note
     * @param mixed $appver
     */
    public function setDevice($device, $deviceid, $status, $os, $ip, $note, $appver)
    {
        $prop = [];
        $attr = "seciossdeviceid;x-dev-$device";
        $devids = $this->_to_array(isset($this->prop[$attr]) ? $this->prop[$attr] : []);
        $match = false;
        for ($i = 0; $i < count($devids); $i++) {
            if (preg_match('/^'.$deviceid.'#([^=]+)=/', $devids[$i], $matches)) {
                $current_status = $matches[1];
                if (strcmp('active', $current_status) === 0) {
                    $status = 'active';
                } elseif (strcmp('inactive', $current_status) === 0) {
                    return PEAR::raiseError('Device is inactive', AUTO_LOGIN_DEVICE_INACTIVE);
                }
                $devids[$i] = preg_replace("/^$deviceid#(active|pending)=.*$/i", "$deviceid#$status=$os+".date('Y/m/d H:i:s')."+$ip###$appver", $devids[$i]);
                $match = true;
                break;
            }
        }
        if (!$match) {
            array_push($devids, $deviceid."#$status=$os+".date('Y/m/d H:i:s')."+$ip###$appver");
        }
        $prop[$attr] = $devids;
        $rc = $this->setProp($prop);
        if (PEAR::isError($rc)) {
            return $rc;
        } elseif ($match) {
            return PEAR::raiseError('Device has been updated', AUTO_LOGIN_IN_HISTORY);
        }
        return $rc;
    }

    public function deleteDevice($device, $deviceid)
    {
        $prop = [];
        $attr = "seciossdeviceid;x-dev-$device";
        $devids = $this->_to_array(isset($this->prop[$attr]) ? $this->prop[$attr] : []);
        $match = false;
        for ($i = 0; $i < count($devids); $i++) {
            if (preg_match('/^'.$deviceid.'#([^=]+)=/', $devids[$i], $matches)) {
                $match = true;
                unset($devids[$i]);
                continue;
            }
        }
        if (!$match) {
            return PEAR::raiseError('Device id is not exist', AUTO_LOGIN_DEVICE_NOEXIST);
        }
        $prop[$attr] = $devids;
        $rc = $this->setProp($prop);
        if (PEAR::isError($rc)) {
            return $rc;
        }
    }

    // {{{ setSecret()

    /**
     *  added by okazaki 20090710
     *  シークレットを暗号化してLDAPに格納する
     *
     *  @access public
     *
     *  @param  string   シークレット
     * @param mixed      $secret
     * @param null|mixed $pin
     * @param null|mixed $deviceid
     * @param null|mixed $device
     * @param null|mixed $otplen
     * @param null|mixed $timewindow
     * @param null|mixed $os
     * @param null|mixed $ip
     *
     *  @return mixed    true:正常終了 PEAR_Error:エラー
     */
    public function setSecret($secret, $pin = null, $deviceid = null, $device = null, $otplen = null, $timewindow = null, $os = null, $ip = null)
    {
        $secret_exists = false;
        $agent = '';

        if (!$secret || strlen($secret) > 255) {
            return PEAR::raiseError('Invalid secret', AUTO_LOGIN_INVALID_VALUE);
        }

        if ($device && !$deviceid) {
            return PEAR::raiseError('No device id', AUTO_LOGIN_INVALID_VALUE);
        }

        if ($deviceid) {
            $agent = 'computer';
            if (preg_match('/iPad/', $_SERVER['HTTP_USER_AGENT'])) {
                $agent = 'ipad';
            } elseif (preg_match('/iPhone/', $_SERVER['HTTP_USER_AGENT'])) {
                $agent = 'iphone';
            } elseif (preg_match('/Android/', $_SERVER['HTTP_USER_AGENT'])) {
                $agent = 'android';
            }
            if ($os) {
                if (preg_match('/Windows/', $os)) {
                    $agent = 'windows';
                } elseif (preg_match('/Mac/', $os)) {
                    $agent = 'mac';
                } elseif (preg_match('/Linux/', $os)) {
                    $agent = 'linux';
                }
            }
            /* Users can register same device id
            if ($agent != 'ipad' && $agent != 'iphone') {
                $username = $this->prop['seciossloginid'];
                $tenant = '';
                if (preg_match('/^[^@]+@(.*)$/', $username, $matches)) {
                    $tenant = $matches[1];
                    $tenant = preg_replace('/([*()])/', '\\$1', $tenant);
                }
                $filter = "(&(!(".$this->options['userattr']."=$username))(&(".$this->options['userattr']."=*@$tenant)(seciossdeviceid;x-dev-$device=$deviceid#*)))";
                if ($this->options['userfilter']) {
                    $filter = '(&'.$filter.$this->options['userfilter'].')';
                }
                $res = @ldap_search($this->conn, $this->options['basedn'], $filter);
                if ($res == false) {
                    return PEAR::raiseError(ldap_error($this->conn), 1);
                }
                $num = ldap_count_entries($this->conn, $res);
                if ($num) {
                    return PEAR::raiseError("Secret is already registered", AUTO_LOGIN_ALREADY_EXISTS);
                }
            }
            */
        }

        $prop = [];
        if (array_search('seciossOtpUser', $this->prop['objectclass']) === false) {
            $prop['objectclass'] = $this->prop['objectclass'];
            array_push($prop['objectclass'], 'seciossOtpUser');
        }

        $encrypt = $this->encrypt($secret);
        if (PEAR::isError($encrypt)) {
            return $encrypt;
        }
        if ($device) {
            $attr = $this->options['secretattr'].";x-dev-$device";
            $hashid = md5($deviceid);
            $secrets = isset($this->prop[$attr]) ? $this->_to_array($this->prop[$attr]) : [];
            $match = false;
            for ($i = 0; $i < count($secrets); $i++) {
                if (preg_match("/^$hashid#/", $secrets[$i])) {
                    $secrets[$i] = $hashid.'#'.$encrypt;
                    $match = true;
                    break;
                }
            }
            if ($match) {
                $secret_exists;
            } else {
                array_push($secrets, $hashid.'#'.$encrypt);
            }
            $prop[$attr] = $secrets;
        } else {
            $prop[$this->options['secretattr']] = $encrypt;
        }

        if ($pin) {
            $encrypt = $this->encrypt($pin);
            if (PEAR::isError($encrypt)) {
                return $encrypt;
            }
            if ($device) {
                $attr = "seciossotpsecretpin;x-dev-$device";
                $hashid = md5($deviceid);
                $pins = isset($this->prop[$attr]) ? $this->_to_array($this->prop[$attr]) : [];
                $match = false;
                for ($i = 0; $i < count($pins); $i++) {
                    if (preg_match("/^$hashid#/", $pins[$i])) {
                        $pins[$i] = $hashid.'#'.$encrypt;
                        $match = true;
                        break;
                    }
                }
                if (!$match) {
                    array_push($pins, $hashid.'#'.$encrypt);
                }
                $prop[$attr] = $pins;
            } else {
                $prop['seciossotpsecretpin'] = $encrypt;
            }
        }
        if ($deviceid) {
            $attr = "seciossdeviceid;x-dev-$device";
            $devids = $this->_to_array(isset($this->prop[$attr]) ? $this->prop[$attr] : []);
            $match = false;
            for ($i = 0; $i < count($devids); $i++) {
                if (preg_match('/^'.$deviceid.'#/', $devids[$i])) {
                    $devids[$i] = preg_replace("/^$deviceid#(active|inactive)=.*$/i", "$deviceid#\\1=$agent+".date('Y/m/d H:i:s')."+$ip", $devids[$i]);
                    $match = true;
                    break;
                }
            }
            if (!$match) {
                array_push($devids, $deviceid."#inactive=$agent+".date('Y/m/d H:i:s')."+$ip");
            }
            $prop[$attr] = $devids;
        }
        if ($otplen) {
            $prop['seciossotpoption'] = "otplen=$otplen";
        }
        if ($timewindow) {
            if (isset($prop['seciossotpoption'])) {
                $prop['seciossotpoption'] .= "#timewindow=$timewindow";
            } else {
                $prop['seciossotpoption'] = "timewindow=$timewindow";
            }
            $file = null;
            if (isset($_GET['app'])) {
                $file = $_GET['app'];
            } elseif (preg_match('/\/(secret|qrsecret|websecret)\.php$/', $_SERVER['SCRIPT_NAME'], $matches)) {
                $file = $matches[1];
            }
            $type = null;
            switch ($file) {
                case 'secret':
                    $type = 'hard';
                    break;
                case 'qrsecret':
                    if (isset($_GET['st'])) {
                        $type = 'soft';
                    } else {
                        $type = 'slinkpass';
                    }
                    break;
                case 'websecret':
                    $type = 'web';
                    break;
            }
            if ($type) {
                $prop['seciossotpoption'] .= "#type=$type";
            }
        }

        $rc = $this->setProp($prop);
        if (PEAR::isError($rc)) {
            return $rc;
        } elseif ($secret_exists) {
            return PEAR::raiseError('Secret has been changed', AUTO_LOGIN_IN_HISTORY);
        }
        return $rc;
    }

    // }}}

    // {{{ getSecret()

    /**
     *  added by okazaki 20090710
     *  暗号化されているシークレットを復号化して取得する
     *
     *  @access public
     *
     * @param null|mixed $app
     *
     *  @return string   シークレット
     */
    public function getSecret($app = null)
    {
        $secret = '';

        if (isset($this->prop[$this->options['secretattr']])) {
            $encrypt = $this->prop[$this->options['secretattr']];
            $secret = $this->decrypt($encrypt);
        }

        return $secret;
    }

    // }}}

    public function _rest_modify($dn, $info, $op = 'replace')
    {
        if ($this->options['api_basedn']) {
            $dn = preg_replace('/,'.$this->options['basedn'].'$/i', ','.$this->options['api_basedn'], $dn);
        }
        $info = [$op => $info];
        $url = $this->options['api_url'].(strpos($this->options['api_url'], '?') === false ? '?' : '&').'action=modify&user='.($this->euser ? $this->euser : $dn).'&dn='.urlencode($dn).(isset($_POST['requestid']) ? '&requestid='.$_POST['requestid'] : '');
        $req = new HTTP_Request($url);
        $req->setMethod(HTTP_REQUEST_METHOD_POST);
        $req->addHeader('Content-type', 'application/json');
        if ($info) {
            $req->setBody(json_encode($info));
        }

        $error = null;
        for ($i = 0; $i < 5; $i++) {
            $res = $req->sendRequest();
            if (PEAR::isError($res)) {
                $error = 'Send request failure('.$res->getMessage().')';
                continue;
            }

            $response = $req->getResponseCode();
            if ($response >= 300) {
                $error = "Bad response code($response)";
            } elseif ($response) {
                $error = null;
                break;
            }
        }
        if ($error) {
            return PEAR::raiseError($error, 52);
        }

        $r = json_decode($req->getResponseBody());
        if ($r->code) {
            return PEAR::raiseError($r->message, $r->code);
        }

        return 0;
    }
}
